#!/bin/sh

# Comprehensive testing system for CEC
# Runs the parser, module expander, dismantler, GRC generator, dependency
# generator, GRC optimizer, PDG generator, and circuit generator.
#
# Most results are compared to reference files, and many are run through
# either the GRC or BLIF simulator before checking.

STRLXML=./cec-strlxml
XMLSTRL=./cec-xmlstrl
EXPAND=./cec-expandmodules
DISMANTLE=./cec-dismantle
ASTGRC=./cec-astgrc
GRCSIM=./cec-grcsim
GRCDOT=./cec-grcdot
GRCOPT=./cec-grcopt
GRCPDG=./cec-grcpdg
PDGBLIF=./cec-pdgblif
SMBLIF=./cec-smblif
BLIFSIM=./cec-blifutil
BLIFDOT="./cec-blifutil -d"

globallog=testall.log
rm -f $globallog
error=0
globalerror=0

keep=0
dot=0
ps=0

Usage() {
    echo "Usage: testall.sh [options] [.strl files]"
    echo "-k    Keep intermediate files"
    echo "-d    Generate .dot files for intermediate representations"
    echo "-p    Generate .ps files for intermediate representations"
    echo "-h    Print this help"
    exit 1
}

# echo requires some testing to see whether it accepts the -n option
case `echo "testing\c"; echo 1,2,3`,`echo -n testing; echo 1,2,3` in
  *c*,-n*) ECHO_N= ECHO_C='
' ECHO_T='      ' ;;
  *c*,*  ) ECHO_N=-n ECHO_C= ECHO_T= ;;
  *)       ECHO_N= ECHO_C='\c' ECHO_T= ;;
esac

# Compare <outfile> <reffile1> <reffile2> <difffile>
# Compares the outfile with reffile1 if reffile1 exists, otherwise,
# compares the outfile with reffile2.  Differences, if any, written to difffile
Compare() {
    if [ $error -eq 0 ] ; then
	generatedfiles="$generatedfiles $4"
	if [ -f "$2" ] ; then
	    echo diff -b $1 $2 ">" $4 1>&2
	    diff -b "$1" "$2" > "$4" 2>&1 || {
		echo "FAILED: $1 differs"
		echo "FAILED $1 differs from $2" 1>&2
		error=1
	    }
	elif [ -f "$3" ] ; then
	    echo diff -b $1 $3 ">" $4 1>&2
	    diff -b "$1" "$3" > "$4" 2>&1 || {
		echo "FAILED: $1 differs"
		echo "FAILED $1 differs from $3" 1>&2
		error=1
	    }
	else
	    echo "FAILED: $3 missing" 1>&2
	    echo "FAILED: $3 missing"
	    error=1
	fi
    fi
}

# Run <args>
# Report the command, run it, and report any errors
Run() {
    if [ $error -eq 0 ] ; then
	echo $* 1>&2
	eval $* 
	error=$?
	if [ $error -ne 0 ]; then
	    if [ $error -ne 100 ]; then
		echo "FAILED: $1 terminated"
		error=1
	    fi
	fi
    fi
}

# RunPass <executable> <infile> <outfile>
# Remembers that the <outfile> was generated
RunPass() {
    generatedfiles="$generatedfiles $3"
    Run $1 "<" $2 ">" $3
}

# RunGRCSIM <refname> <basename> <extension>
RunGRCSIM() {
    generatedfiles="$generatedfiles $2.$3.simout"
    if [ -f $1.tv ] ; then
	Run $GRCSIM -t $1.tv "<" $2.$3.xml ">" $2.$3.simout
    else
	Run $GRCSIM -c 10 "<" $2.$3.xml ">" $2.$3.simout
    fi

    if [ $error -eq 100 ]; then
	echo $ECHO_N "($3 cyclic) $ECHO_C"
	error=0
    else
	Compare $2.$3.simout $1.simout $1.simout \
	    $2.$3.sim.diff
    fi
}

Check() {
    error=0
    basename=`echo $1 | sed 's/.*\\///
                             s/.strl//'`
    reffile=`echo $1 | sed 's/.strl$//'`

    echo $ECHO_N "Testing $basename...$ECHO_C"

    echo 1>&2
    echo "###### Testing $basename" 1>&2

    generatedfiles=

    # Test the parser

    RunPass $STRLXML $1 ${basename}.ast.xml
    RunPass $XMLSTRL ${basename}.ast.xml ${basename}.ast.strl
    Compare ${basename}.ast.strl ${reffile}.out $1 ${basename}.ast.diff

    # Test the expander

    RunPass $EXPAND ${basename}.ast.xml ${basename}.exp.xml
    RunPass $XMLSTRL ${basename}.exp.xml ${basename}.exp.strl
    Compare ${basename}.exp.strl ${reffile}.expout $1 ${basename}.exp.diff

    # Test the dismantler
   
    RunPass $DISMANTLE ${basename}.exp.xml ${basename}.dis.xml
    RunPass $XMLSTRL ${basename}.dis.xml ${basename}.dis.strl
    Compare ${basename}.dis.strl ${reffile}.disout $1 ${basename}.dis.diff

    # Test GRC generation

    RunPass $ASTGRC ${basename}.dis.xml ${basename}.grc.xml

    # Generate a .dot file if requested

    if [ $dot -ne 0 ]; then
	RunPass $GRCDOT ${basename}.grc.xml ${basename}.grc.dot
	if [ $ps -ne 0 ]; then
	    Run dot -Tps -o ${basename}.grc.ps ${basename}.grc.dot
	fi
    fi

    # Run the GRC simulator

    if [ -f ${reffile}.stop-before-simulation ] ; then
	echo $ECHO_N "(partial) $ECHO_C"
	echo "##### Stopped here because of ${reffile}.stop-before-simulation" 1>&2
	error=2
    fi

    RunGRCSIM ${reffile} ${basename} grc

    # Run the GRC optimizer

    RunPass "$GRCOPT" ${basename}.grc.xml ${basename}.opt.xml

    # Generate a .dot file if requested

    if [ $dot -ne 0 ]; then
	RunPass $GRCDOT ${basename}.opt.xml ${basename}.opt.dot
	if [ $ps -ne 0 ]; then
	    Run dot -Tps -o ${basename}.opt.ps ${basename}.opt.dot
	fi
    fi

    # Run the GRC simulator

    RunGRCSIM ${reffile} ${basename} opt

    # Translate into PDG

    RunPass "$GRCPDG" ${basename}.opt.xml ${basename}.pdg.xml

    # Generate .dot file if requested

    if [ $dot -ne 0 ]; then
	RunPass $GRCDOT ${basename}.pdg.xml ${basename}.pdg.dot
	if [ $ps -ne 0 ]; then
	    Run dot -Tps -o ${basename}.pdg.ps ${basename}.pdg.dot
	fi
    fi

    # Run the GRC simulator and check the results

    RunGRCSIM ${reffile} ${basename} pdg

    # Run the circuit synthesizer

    # Stop here if requested
    if [ -f ${reffile}.stop-before-circuit ] ; then
	echo $ECHO_N "(partial) $ECHO_C"
	echo "##### Stopped here because of ${reffile}.stop-before-circuit" 1>&2
	error=2
    fi

    generatedfiles="$generatedfiles ${basename}.core.blif ${basename}.sm"
    Run $PDGBLIF --sm ${basename}.sm "<" ${basename}.pdg.xml \
        ">" ${basename}.core.blif

    # Convert the trivial state machine file into BLIF

    generatedfiles="$generatedfiles ${basename}.sm.blif ${basename}.out.blif"
    Run $SMBLIF ${basename}.sm ${basename}.sm.blif
    Run sed '/^[.]end/d' ${basename}.core.blif ">" ${basename}.out.blif
    Run cat ${basename}.sm.blif ">>" ${basename}.out.blif
    Run echo ".end" ">>" ${basename}.out.blif

    # Generate .dot file if requested

    if [ $dot -ne 0 ]; then
	RunPass "$BLIFDOT" ${basename}.out.blif ${basename}.blif.dot
	if [ $ps -ne 0 ]; then
	    Run dot -Tps -o ${basename}.blif.ps ${basename}.blif.dot
	fi
    fi

    # Run the circuit simulator and compare the results

    if [ $error -eq 0 ]; then
	generatedfiles="$generatedfiles ${basename}.blif.simout ${basename}.blif.sim.diff"
	if [ -f ${reffile}.tv ] ; then
	    # Use a test vector file
	    Run $BLIFSIM -t ${reffile}.tv ${basename}.out.blif \
		">" ${basename}.blif.simout
	else
            # Run a certain number of cycles
	    Run $BLIFSIM -c `cat ${reffile}.simout | wc -l` \
		${basename}.out.blif ">" ${basename}.blif.simout
	fi
	Compare ${basename}.blif.simout ${reffile}.simout  ${reffile}.simout \
	    ${basename}.blif.sim.diff
    fi

    # Report the status and clean up the generated files

    if [ $error -ne 1 ]; then
	if [ $keep -eq 0 ]; then
	    rm -f $generatedfiles
	fi
	generatedfiles=
	echo "OK"
	if [ $error -eq 0 ] ; then
	    echo "###### SUCCESS" 1>&2
	else
	    echo "###### PARTIAL SUCCESS" 1>&2
	fi
	error=0
    else
	echo "###### FAILED" 1>&2
	globalerror=$error
	error=0
    fi

}

while getopts kdph c; do
    case $c in
	k) # Keep intermediate files
	    keep=1
	    ;;
	d) # Generate dot files
	    keep=1
	    dot=1
	    ;;
	p) # Generate Postscript files
	    keep=1
	    dot=1
	    ps=1
	    ;;
	h) # Help
	    Usage
	    ;;
    esac
done

shift `expr $OPTIND - 1`

if [ $# -ge 1 ]
then
    files=$@
else
    files=tests/test*.strl
fi

for file in $files
do
  Check $file 2>> $globallog
done

exit $globalerror
